package netserv.apps.activestreaming.server;

import java.io.*;
import java.net.*;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;
import java.util.StringTokenizer;
import java.util.logging.Logger;

import javax.servlet.ServletConfig;
import javax.servlet.ServletException;
import javax.servlet.http.*;

import org.mortbay.jetty.*;
import org.mortbay.jetty.servlet.*;

/**
 * Content Server : All the request comes to this Servlet. This Servlet
 * redirects all the request to NetServ nodes, if any. otherwise creates a
 * direct connection to streaming server.
 * 
 * @author Aman
 * 
 */
public class ContentServer extends HttpServlet {

	private static final long serialVersionUID = 1L;
	// Map for maintaining current streaming files.
	private Map<String, VideoStreamer> streamingMap = new HashMap<String, VideoStreamer>();
	private static String SERVER_IP;
	private static String STREAM_SERVER_IP;
	private static String NETSERV_NODE_IP;
	private static final String CONTENT_SERVER_PORT = "8088";
	private static final String NETSERV_NODE_PORT = "8888";
	private static final String STREAM_SERVER_PORT = "8080";
	public static final String LOCALROOT = "./sample";

	private static final Logger log = Logger.getLogger(ContentServer.class
			.getName());

	// NetServ Configuration
	static final String SERVER_PROPERTIES = "./server.properties";

	// These static variables are also being set in the server.properties file
	public static String WEBROOT = "./";
	public static String NSIS_TRIGGER = "./netserv-trigger";
	public static String GEOIP_FILE = "./GeoLiteCity.dat";
	public static int MODULE_LIFETIME = 24000;
	public static double THRESHOLD_DISTANCE = 1000.0; // kilometers ?

	ContentSingleton singleton = ContentSingleton.getInstance();

	/**
	 * Servlet context initializer
	 */
	public void init(ServletConfig config) throws ServletException {
		super.init(config);
		singleton.readTranslationToHashMap();
	}

	public void doGet(HttpServletRequest request, HttpServletResponse response) {
		String file = request.getParameter("file");
		String mode = request.getParameter("mode");
		final String client = request.getRemoteAddr();

		log.info("Request for " + file + " from " + request.getRemoteAddr());
		if (file == null) {
			try {
				response.getWriter().write("File not found !");
				return;
			} catch (IOException e2) {
				e2.printStackTrace();
			}
		}
		
		// NetServ using NSIS
		// sendNetServSignal(client, file, mode, response);
		// NetServ using simple setup
		simpleSetup(client, file, mode, response);

	}

	private void simpleSetup(final String client, String file, String mode,
			HttpServletResponse response) {
		if (mode != null && mode.equalsIgnoreCase("live")) {
			log.info("simpleSetup : Sending to NetServ Node .. mode=Live");
			sendVLCVideo(file, response, true, false, NETSERV_NODE_IP);
		} else {
			sendVLCVideo(file, response, false, false, NETSERV_NODE_IP);
		}
	}

	private void sendNetServSignal(final String client, String file,
			String mode, HttpServletResponse response) {
		String netServNode = getNetServNode(client);
		if (netServNode == null) {
			// First request
			Thread signalThread = new Thread(new Runnable() {
				public void run() {
					String node = null;
					// 5 times: sleep 2 second and probe the install
					for (int i = 0; i < 5; i++) {
						sendNetServSetup(client);
						try {
							Thread.sleep(2000);
						} catch (InterruptedException e) {
							log.severe("signalThread : Error while sleeping: "
									+ e.toString());
						}
						node = sendNetServProbe(client);
						if (node != null) {
							break;
						}
					}
					try {
						Thread.sleep(MODULE_LIFETIME * 1000 - 5000);
					} catch (InterruptedException e) {
						log.severe("signalThread : Error while sleeping: "
								+ e.toString());
					}

					if (node != null)
						sendNetServTeardown(node);
				}
			});
			signalThread.start();

			if (mode != null && mode.equalsIgnoreCase("live")) {
				log.info("sendNetServSignal : Sending directly to streaming server .. mode=Live");
				sendVLCVideo(file, response, true, true, null);
			} else {
				log.info("sendNetServSignal : Sending directly to streaming server..");
				sendVLCVideo(file, response, false, true, null);
			}
		} else {
			// NetServ node is present
			if (mode != null && mode.equalsIgnoreCase("live")) {
				log.info("sendNetServSignal : Sending to NetServ Node .. mode=Live");
				sendVLCVideo(file, response, true, false, netServNode);
			} else {
				log.info("sendNetServSignal : Sending to NetServ Node..");
				sendVLCVideo(file, response, false, false, netServNode);
			}
		}
	}

	/**
	 * Returns the nearest NetServ Node for a client.
	 * 
	 * @param client
	 * @return
	 */
	private String getNetServNode(String client) {
		// Get all NetServ nodes
		Set<String> nodes = singleton.getNodes();

		String netServNode = null;
		double globalDistance = THRESHOLD_DISTANCE;
		try {
			Iterator<String> it = nodes.iterator();
			log.info("getNetServNode : Total nodes found - " + nodes.size());
			while (it.hasNext()) {
				String node = (String) it.next();
				double distance = singleton.calc_distance(netServNode, client);
				if (distance < globalDistance) {
					globalDistance = distance;
					netServNode = node;
				}
			}
		} catch (Exception e) {
			e.printStackTrace();
			log.severe("getNetServNode : Error calculating NetServ Node distance.");
		}

		if (globalDistance < THRESHOLD_DISTANCE) {
			log.info("getNetServNode : Found NetServ node " + netServNode
					+ " [" + client + ", " + globalDistance + "]");
			return netServNode + ":" + NETSERV_NODE_PORT;
		} else {
			return null;
		}
	}

	/**
	 * Send NetServ setup signal
	 * 
	 * @param client
	 *            - IP address of the client node
	 * @return Boolean value determining whether operation was successful or not
	 */
	private boolean sendNetServSetup(String client) {
		String s;
		String command = NSIS_TRIGGER
				+ " "
				+ client
				+ " -s -user jae -id NetServ.apps.ActiveStreaming_1.0.0 -url http://netserv-server/modules/activestreaming.jar -ttl "
				+ MODULE_LIFETIME;
		log.info("sendNetServSetup : " + command);
		Process p;
		try {
			p = Runtime.getRuntime().exec(command);
			BufferedReader stdInput = new BufferedReader(new InputStreamReader(
					p.getInputStream()));

			while ((s = stdInput.readLine()) != null) {
				// output will look like: 2 0 0
				if (s.equals("2 0 0"))
					return true;
				else
					return false;
			}
		} catch (IOException e) {
			log.severe("sendNetServSetup : Error running trigger setup \n"
					+ e.toString());
		}
		return false;
	}

	/**
	 * Send NetServ probe.
	 */
	private String sendNetServProbe(String client) {
		String command = NSIS_TRIGGER
				+ " "
				+ client
				+ " -p -user jae -id NetServ.apps.ActiveStreaming_1.0.0 -probe 2";
		log.info("sendNetServProbe : " + command);
		BufferedReader stdInput;
		Process p;
		try {
			p = Runtime.getRuntime().exec(command);
			stdInput = new BufferedReader(new InputStreamReader(
					p.getInputStream()));
			// output will look like:
			// 1.2.3.4 ACTIVE (for working nodes)
			// 1.2.3.4 NOT PRESENT (for non-working nodes)
			String s;
			while ((s = stdInput.readLine()) != null) {
				StringTokenizer st = new StringTokenizer(s);
				String ipAddr = null;
				String status = null;
				ipAddr = st.nextToken();
				if (ipAddr != null) {
					status = st.nextToken();
				}
				if (status.equals("ACTIVE")) {
					if (singleton.addNode(ipAddr))
						log.info("sendNetServProbe : Adding NetServ node "
								+ ipAddr);
					return ipAddr;
				}
			}
		} catch (IOException e) {
			log.severe("sendNetServProbe : NSIS trigger not present");
			log.severe("sendNetServProbe : Error adding NetServ node"
					+ e.toString());
		}
		return null;
	}

	/**
	 * Removes the NetServ Node mapping for given IP address.
	 * 
	 * @param ipAddr
	 * @return
	 */

	private boolean sendNetServTeardown(String ipAddr) {
		String command = NSIS_TRIGGER + " " + ipAddr
				+ " -r -user jae -id NetServ.apps.ActiveStreaming_1.0.0";
		log.info("sendNetServTeardown : " + command);
		try {
			Runtime.getRuntime().exec(command);
			singleton.removeNode(ipAddr);
			log.info("sendNetServTeardown : Removing NetServ node " + ipAddr);
			return true;
		} catch (IOException e) {
			log.info("sendNetServTeardown : Error removing node list: "
					+ e.toString());
		}
		return false;
	}

	/**
	 * Returns a VLC plugin embedded HTML page.
	 * 
	 * @param file
	 *            - streaming file name
	 * @param response
	 *            - http response object
	 * @param isLive
	 *            - whether live mode or not
	 * @param isDirect
	 *            - whether there direct connection to streaming server
	 */
	private void sendVLCVideo(String file, HttpServletResponse response,
			boolean isLive, boolean isDirect, String server) {
		String url = null;
		if (isDirect) {
			url = this.directURL(file);
		} else {
			if (isLive) {
				url = this.netServNodeURL(file, server);
				url += "&mode=live";
			} else {
				url = this.netServNodeURL(file, server);
				url += "&mode=vod";
			}
		}
		try {
			PrintWriter pr = response.getWriter();
			pr.write("<html>");
			pr.write("<head><title>NetServ Active Streaming</title></head>");
			pr.write("<body>");
			pr.write("<div id=\"content\" align=\"center\">");
			pr.write("<h2>NetServ Active Streaming</h2>");

			pr.write("<div align=\"center\"><a href=\"http://"
					+ ContentServer.SERVER_IP + ":"
					+ ContentServer.CONTENT_SERVER_PORT + "/stream/?file="
					+ file + "&mode=live\">View Live</a> </div>");

			pr.write("<embed type=\"application/x-vlc-plugin\" name="
					+ file
					+ " autoplay=\"yes\" loop=\"no\" width=\"680\" height=\"460\" target=\""
					+ url + "\"" + " />");
			pr.write("</div>");
			pr.write("<br />");
			pr.write("</body>");
			pr.write("</html>");
			pr.flush();
		} catch (IOException e1) {
			e1.printStackTrace();
		}
	}

	/**
	 * Used to send HTML5 video instead of VLC plugin. Tested it on Chrome 10
	 * 
	 * @param file
	 * @param response
	 */
	private void sendHTML5Video(String file, HttpServletResponse response,
			String server) {
		try {
			PrintWriter pr = response.getWriter();
			pr.write("<html>");
			pr.write("<head><title>NetServ Active Streaming</title></head>");
			pr.write("<body>");
			pr.write("<div id=\"content\" align=\"center\">");
			pr.write("<h2>NetServ Active Streaming</h2>");
			pr.write("<div id=\"netserv-video\">");
			pr.write("<video id=\"demo-video\" controls>");
			pr.write("<source src=\"" + this.netServNodeURL(file, server)
					+ "\"" + "type=\"video/ogg\" />");
			pr.write("</video>");
			pr.write("</div></div>");
			pr.write("<br />");
			pr.write("</body>");
			pr.write("</html>");
			pr.flush();
		} catch (IOException e1) {
			e1.printStackTrace();
		}
	}

	/**
	 * 
	 * @param file
	 * @return
	 */
	private String netServNodeURL(String file, String server) {
		String newurl = "";
		newurl = "http://" + server + ":" + NETSERV_NODE_PORT
				+ "/stream-cdn/?url=" + directURL(file);
		return newurl;
	}

	private String directURL(String file) {
		String newurl = "";
		newurl = "http://" + STREAM_SERVER_IP + ":" + STREAM_SERVER_PORT;
		return newurl;
	}

	/**
	 * Returns the first non looping ipv4 address.
	 * 
	 * @return
	 */
	private static InetAddress getFirstNonLoopbackAddress() {
		Enumeration<NetworkInterface> en = null;
		try {
			en = NetworkInterface.getNetworkInterfaces();
		} catch (SocketException e) {
			log.severe("Error getting network interface.." + e.toString());
		}
		while (en.hasMoreElements()) {
			NetworkInterface i = (NetworkInterface) en.nextElement();
			for (Enumeration en2 = i.getInetAddresses(); en2.hasMoreElements();) {
				InetAddress addr = (InetAddress) en2.nextElement();
				if (!addr.isLoopbackAddress()) {
					if (addr instanceof Inet4Address) {
						return addr;
					}
				}
			}
		}
		return null;
	}

	/**
	 * Only on Linux systems, embedded VLC streaming server is supported media
	 * files are streamed on - http://ipaddress:8080/stream/file_1. mp4 - "asf"
	 * mux container 2. ogv - "ogg" mux container
	 * 
	 * @param args
	 * @throws Exception
	 */
	public static void main(String[] args) throws Exception {
		// starting streaming service
		if (args.length == 2) {
			String file = args[0];
			SERVER_IP = args[1];
			// netserv node ip is calculated dynamically by NSIS
			NETSERV_NODE_IP = SERVER_IP;
			STREAM_SERVER_IP = SERVER_IP;
			log.info("Content Server address : " + SERVER_IP);
			// VideoStreamer.playMedia(LOCALROOT + "/" + file, file,
			// "asf",SERVER_IP, STREAM_SERVER_PORT);
			// starting jetty server
			Server server = new Server(
					Integer.parseInt(ContentServer.CONTENT_SERVER_PORT));
			Context root = new Context(server, "/", Context.SESSIONS);
			root.addServlet(new ServletHolder(new ContentServer()), "/stream/*");
			log.info("Content Server started..");
			server.start();
			server.join();
		} else {
			log.severe("Usage: ContentServer <movie_file_name> <content server ip> <netserv node ip>");
		}

	}
}